#include <winsock.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#define VERSION 0x0101
#define PRINCIPAL HIBYTE(VERSION)
#define RELEASE   LOBYTE(VERSION)
#define NUM_MIN_SOCKETS 3

#define PUERTO 12345

int main(void)
{
	WSADATA wsaData;
	SOCKET UDP_srv_sock, TCP_listen_sock, TCP_msg_sock;
	struct sockaddr_in local, from;
	fd_set lectura;
	struct timeval timeout;
	int fromlen;
	char Buffer[128];
	int select_val;
	int retval;
	
	// socket servidor protocolo UDP
	UDP_srv_sock=INVALID_SOCKET;
	// socket servidor protocolo TCP
	TCP_listen_sock=INVALID_SOCKET;
	// socket para atender clientes TCP
	TCP_msg_sock=INVALID_SOCKET;
	
	
	// Inicializar la libreria de sockets
	if (WSAStartup(VERSION,&wsaData) == SOCKET_ERROR) {
		fprintf(stderr,"fallo en WSAStartup con error %d\n",WSAGetLastError());
		WSACleanup();
		return -1;
	}
	
	// definimos la direccion del puerto del servidor	
	local.sin_family = AF_INET;
	local.sin_addr.s_addr = INADDR_ANY;
	local.sin_port=htons(PUERTO);
	
	// creamos los sockets que van a estar a la escucha
	TCP_listen_sock = socket(AF_INET, SOCK_STREAM,0); // TCP socket
	if (TCP_listen_sock == INVALID_SOCKET)
	{
		fprintf(stderr,"fallo en socket() al crear servidor TCP con error %d\n",WSAGetLastError());
		WSACleanup();
		return -1;
	}
	
	UDP_srv_sock = socket(AF_INET, SOCK_DGRAM,0); // TCP socket
	if (UDP_srv_sock == INVALID_SOCKET)
	{
		fprintf(stderr,"fallo en socket() al crear servidor UDP con error %d\n",WSAGetLastError());
		WSACleanup();
		return -1;
	}
	
	// nombramos los servidores	
	if (bind(TCP_listen_sock,(struct sockaddr*)&local,sizeof(struct sockaddr_in))==SOCKET_ERROR)
	{
		fprintf(stderr,"fallo en bind() al nombrar socket TCP con error %d\n",WSAGetLastError());
		WSACleanup();
		return -1;
	}
	if (bind(UDP_srv_sock,(struct sockaddr*)&local,sizeof(struct sockaddr_in))==SOCKET_ERROR)
	{
		fprintf(stderr,"fallo en bind() al nombrar socket UDP con error %d\n",WSAGetLastError());
		WSACleanup();
		return -1;
	}
	
	
	// Poner el socket TCP a la escucha
	
	if (listen(TCP_listen_sock,5) == SOCKET_ERROR)
	{
		fprintf(stderr,"fallo en listen() con error %d\n",WSAGetLastError());
		WSACleanup();
		return -1;
	}
	
	// usaremos una temporizacion para select de 1 segundo
	timeout.tv_sec=1;
	timeout.tv_usec=0;
	
	// Bucle principal del servidor (es un bucle indefinido)
	while(1)
	{
		fromlen =sizeof(struct sockaddr_in);
		
		
		// Detectar que socket tiene datos
		FD_ZERO(&lectura);
		FD_SET(UDP_srv_sock,&lectura);
		FD_SET(TCP_listen_sock,&lectura);
		if (TCP_msg_sock!=INVALID_SOCKET)
			FD_SET(TCP_msg_sock,&lectura);
		
		select_val=select(3,&lectura,NULL,NULL,&timeout);
		if (select_val==SOCKET_ERROR)
		{
			fprintf(stderr,"fallo en select() con error %d\n",WSAGetLastError());
			WSACleanup();
			return -1;
		}
		
		if (select_val>0)
		{
			// Servidor TCP
			if (FD_ISSET(TCP_listen_sock,&lectura))
			{
				// datos pendientes en este socket indican un intento de conexion pendiente
				if (TCP_msg_sock!=INVALID_SOCKET)
				{
					fromlen =sizeof(struct sockaddr_in);
					TCP_msg_sock = accept(TCP_listen_sock,(struct sockaddr*)&from, &fromlen);
					if (TCP_msg_sock == INVALID_SOCKET)
					{
						fprintf(stderr,"fallo en accept() con error %d\n",WSAGetLastError());
						WSACleanup();
						return 1;
					}
					printf("aceptada conexion connection de %s, puerto %d\n",inet_ntoa(from.sin_addr),htons(from.sin_port));
				}
				else
					printf("conexion pendiente\n");
			}
			
			// Servidor UDP
			if (FD_ISSET(UDP_srv_sock,&lectura))
			{
				fromlen =sizeof(struct sockaddr_in);
				retval = recvfrom(UDP_srv_sock,Buffer,sizeof (Buffer),0,(struct sockaddr *)&from,&fromlen);
				printf("Recibido datagrama de %s, puerto %d\n",inet_ntoa(from.sin_addr),htons(from.sin_port));
				
				if (retval == SOCKET_ERROR)
				{
					fprintf(stderr,"fallo en recvfrom() con error %d\n",WSAGetLastError());
				}
				// OJO: el protocolo es muy simple y se envia el \0 final para terminar la cadena
				printf("Recibidos %d bytes, msg: [%s] del cliente UDP\n",retval,Buffer);
				
				printf("Devolviendo los datos al cliente\n");
				retval = sendto(UDP_srv_sock,Buffer,strlen(Buffer),0,(struct sockaddr *)&from,fromlen);
				if (retval == SOCKET_ERROR)
				{
					fprintf(stderr,"fallo en sendto() con error %d\n",WSAGetLastError());
				}
			}
			
			// Conexion con cliente TCP
			if (FD_ISSET(TCP_msg_sock,&lectura))
			{
				retval = recv(TCP_msg_sock,Buffer,sizeof (Buffer),0 );

				if (retval == SOCKET_ERROR)
				{
					fprintf(stderr,"fallo en recv() con error %d\n",WSAGetLastError());
					closesocket(TCP_msg_sock);
					TCP_msg_sock=INVALID_SOCKET;
				}
				if (retval == 0)
				{
					printf("El cliente cerro la conexion\n");
					closesocket(TCP_msg_sock);
					TCP_msg_sock=INVALID_SOCKET;
				}
				else
				{
					printf("Recibidos %d bytes, msg: [%s] del cliente TCP\n",retval,Buffer);
				
					printf("Devolviendo datos al cliente\n");
					retval = send(TCP_msg_sock,Buffer,strlen(Buffer),0);
					if (retval == SOCKET_ERROR)
					{
						fprintf(stderr,"fallo en send() con error %d\n",WSAGetLastError());
					}
					else
					{
						printf("Cerrando la conexion\n");
						closesocket(TCP_msg_sock);
						TCP_msg_sock=INVALID_SOCKET;
					}
				}
			}
		} // fin select_val>0
	} // fin while
}
